@@ -1,4 +1,5 @@
 import errno
+import functools
 import io
 import os
 import stat
@@ -9,31 +10,49 @@ import time
 from collections import defaultdict
 from signal import SIGINT
 
-import llfuse
+from .fuse_impl import llfuse, has_pyfuse3
+
+
+if has_pyfuse3:
+    import trio
+
+    def async_wrapper(fn):
+        @functools.wraps(fn)
+        async def wrapper(*args, **kwargs):
+            return fn(*args, **kwargs)
+        return wrapper
+else:
+    trio = None
+
+    def async_wrapper(fn):
+        return fn
+
 
 from .logger import create_logger
 logger = create_logger()
 
 from .crypto.low_level import blake2b_128
 from .archiver import Archiver
-from .archive import Archive
+from .archive import Archive, get_item_uid_gid
 from .hashindex import FuseVersionsIndex
-from .helpers import daemonize, hardlinkable, signal_handler, format_file_size, Error
+from .helpers import daemonize, daemonizing, hardlinkable, signal_handler, format_file_size
 from .helpers import msgpack
 from .item import Item
 from .lrucache import LRUCache
+from .platform import uid2user, gid2group
 from .remote import RemoteRepository
 
-# Does this version of llfuse support ns precision?
-have_fuse_xtime_ns = hasattr(llfuse.EntryAttributes, 'st_mtime_ns')
-
-# Does this version of llfuse support birthtime?
-have_fuse_birthtime = hasattr(llfuse.EntryAttributes, 'st_birthtime')  # never?
-have_fuse_birthtime_ns = hasattr(llfuse.EntryAttributes, 'st_birthtime_ns')  # since llfuse 1.3
-
 
 def fuse_main():
-    return llfuse.main(workers=1)
+    if has_pyfuse3:
+        try:
+            trio.run(llfuse.main)
+        except:
+            return 1  # TODO return signal number if it was killed by signal
+        else:
+            return None
+    else:
+        return llfuse.main(workers=1)
 
 
 # size of some LRUCaches (1 element per simultaneously open file)
@@ -131,7 +150,7 @@ class ItemCache:
         stream_offset = 0
         # Offset of the current chunk in the metadata stream
         chunk_begin = 0
-        # Length of the chunk preciding the current chunk
+        # Length of the chunk preceding the current chunk
         last_chunk_length = 0
         msgpacked_bytes = b''
 
@@ -139,13 +158,6 @@ class ItemCache:
         meta = self.meta
         pack_indirect_into = self.indirect_entry_struct.pack_into
 
-        def write_bytes(append_msgpacked_bytes):
-            # XXX: Future versions of msgpack include an Unpacker.tell() method that provides this for free.
-            nonlocal msgpacked_bytes
-            nonlocal stream_offset
-            msgpacked_bytes += append_msgpacked_bytes
-            stream_offset += len(append_msgpacked_bytes)
-
         for key, (csize, data) in zip(archive_item_ids, self.decrypted_repository.get_many(archive_item_ids)):
             # Store the chunk ID in the meta-array
             if write_offset + 32 >= len(meta):
@@ -154,16 +166,25 @@ class ItemCache:
             current_id_offset = write_offset
             write_offset += 32
 
-            # The chunk boundaries cannot be tracked through write_bytes, because the unpack state machine
-            # *can* and *will* consume partial items, so calls to write_bytes are unrelated to chunk boundaries.
             chunk_begin += last_chunk_length
             last_chunk_length = len(data)
 
             unpacker.feed(data)
             while True:
                 try:
-                    item = unpacker.unpack(write_bytes)
+                    item = unpacker.unpack()
+                    need_more_data = False
                 except msgpack.OutOfData:
+                    need_more_data = True
+
+                start = stream_offset - chunk_begin
+                # tell() is not helpful for the need_more_data case, but we know it is the remainder
+                # of the data in that case. in the other case, tell() works as expected.
+                length = (len(data) - start) if need_more_data else (unpacker.tell() - stream_offset)
+                msgpacked_bytes += data[start:start+length]
+                stream_offset += length
+
+                if need_more_data:
                     # Need more data, feed the next chunk
                     break
 
@@ -213,32 +234,24 @@ class ItemCache:
         self.write_offset = write_offset
 
 
-class FuseOperations(llfuse.Operations):
-    """Export archive as a FUSE filesystem
+class FuseBackend(object):
+    """Virtual filesystem based on archive(s) to provide information to fuse
     """
-    # mount options
-    allow_damaged_files = False
-    versions = False
-    uid_forced = None
-    gid_forced = None
-    umask = 0
 
-    def __init__(self, key, repository, manifest, args, decrypted_repository):
-        super().__init__()
+    def __init__(self, key, manifest, repository, args, decrypted_repository):
         self.repository_uncached = repository
-        self.decrypted_repository = decrypted_repository
-        self.args = args
-        self.manifest = manifest
+        self._args = args
+        self.numeric_ids = args.numeric_ids
+        self._manifest = manifest
         self.key = key
-        # Maps inode numbers to Item instances. This is used for synthetic inodes,
-        # i.e. file-system objects that are made up by FuseOperations and are not contained
-        # in the archives. For example archive directories or intermediate directories
+        # Maps inode numbers to Item instances. This is used for synthetic inodes, i.e. file-system objects that are
+        # made up and are not contained in the archives. For example archive directories or intermediate directories
         # not contained in archives.
-        self.items = {}
+        self._items = {}
         # cache up to <FILES> Items
         self._inode_cache = LRUCache(capacity=FILES, dispose=lambda _: None)
-        # _inode_count is the current count of synthetic inodes, i.e. those in self.items
-        self._inode_count = 0
+        # _inode_count is the current count of synthetic inodes, i.e. those in self._items
+        self.inode_count = 0
         # Maps inode numbers to the inode number of the parent
         self.parent = {}
         # Maps inode numbers to a dictionary mapping byte directory entry names to their inode numbers,
@@ -247,38 +260,217 @@ class FuseOperations(llfuse.Operations):
         self.default_uid = os.getuid()
         self.default_gid = os.getgid()
         self.default_dir = None
+        # Archives to be loaded when first accessed, mapped by their placeholder inode
         self.pending_archives = {}
         self.cache = ItemCache(decrypted_repository)
-        data_cache_capacity = int(os.environ.get('BORG_MOUNT_DATA_CACHE_ENTRIES', os.cpu_count() or 1))
-        logger.debug('mount data cache capacity: %d chunks', data_cache_capacity)
-        self.data_cache = LRUCache(capacity=data_cache_capacity, dispose=lambda _: None)
-        self._last_pos = LRUCache(capacity=FILES, dispose=lambda _: None)
+        self.allow_damaged_files = False
+        self.versions = False
+        self.uid_forced = None
+        self.gid_forced = None
+        self.umask = 0
 
     def _create_filesystem(self):
         self._create_dir(parent=1)  # first call, create root dir (inode == 1)
-        if self.args.location.archive:
-            if self.versions:
-                raise Error("for versions view, do not specify a single archive, "
-                            "but always give the repository as location.")
-            self.process_archive(self.args.location.archive)
+        if self._args.location.archive:
+            self._process_archive(self._args.location.archive)
         else:
             self.versions_index = FuseVersionsIndex()
-            for archive in self.manifest.archives.list_considering(self.args):
+            for archive in self._manifest.archives.list_considering(self._args):
                 if self.versions:
                     # process archives immediately
-                    self.process_archive(archive.name)
+                    self._process_archive(archive.name)
                 else:
                     # lazily load archives, create archive placeholder inode
                     archive_inode = self._create_dir(parent=1, mtime=int(archive.ts.timestamp() * 1e9))
                     self.contents[1][os.fsencode(archive.name)] = archive_inode
                     self.pending_archives[archive_inode] = archive.name
 
+    def get_item(self, inode):
+        item = self._inode_cache.get(inode)
+        if item is not None:
+            return item
+        try:
+            # this is a cheap get-from-dictionary operation, no need to cache the result.
+            return self._items[inode]
+        except KeyError:
+            # while self.cache does some internal caching, it has still quite some overhead, so we cache the result.
+            item = self.cache.get(inode)
+            self._inode_cache[inode] = item
+            return item
+
+    def check_pending_archive(self, inode):
+        # Check if this is an archive we need to load
+        archive_name = self.pending_archives.pop(inode, None)
+        if archive_name is not None:
+            self._process_archive(archive_name, [os.fsencode(archive_name)])
+
+    def _allocate_inode(self):
+        self.inode_count += 1
+        return self.inode_count
+
+    def _create_dir(self, parent, mtime=None):
+        """Create directory
+        """
+        ino = self._allocate_inode()
+        if mtime is not None:
+            self._items[ino] = Item(internal_dict=self.default_dir.as_dict())
+            self._items[ino].mtime = mtime
+        else:
+            self._items[ino] = self.default_dir
+        self.parent[ino] = parent
+        return ino
+
+    def find_inode(self, path, prefix=[]):
+        segments = prefix + path.split(b'/')
+        inode = 1
+        for segment in segments:
+            inode = self.contents[inode][segment]
+        return inode
+
+    def _process_archive(self, archive_name, prefix=[]):
+        """Build FUSE inode hierarchy from archive metadata
+        """
+        self.file_versions = {}  # for versions mode: original path -> version
+        t0 = time.perf_counter()
+        archive = Archive(self.repository_uncached, self.key, self._manifest, archive_name,
+                          consider_part_files=self._args.consider_part_files)
+        strip_components = self._args.strip_components
+        matcher = Archiver.build_matcher(self._args.patterns, self._args.paths)
+        partial_extract = not matcher.empty() or strip_components
+        hardlink_masters = {} if partial_extract else None
+
+        def peek_and_store_hardlink_masters(item, matched):
+            if (partial_extract and not matched and hardlinkable(item.mode) and
+                    item.get('hardlink_master', True) and 'source' not in item):
+                hardlink_masters[item.get('path')] = (item.get('chunks'), None)
+
+        filter = Archiver.build_filter(matcher, peek_and_store_hardlink_masters, strip_components)
+        for item_inode, item in self.cache.iter_archive_items(archive.metadata.items, filter=filter,
+                                                              consider_part_files=self._args.consider_part_files):
+            if strip_components:
+                item.path = os.sep.join(item.path.split(os.sep)[strip_components:])
+            path = os.fsencode(item.path)
+            is_dir = stat.S_ISDIR(item.mode)
+            if is_dir:
+                try:
+                    # This can happen if an archive was created with a command line like
+                    # $ borg create ... dir1/file dir1
+                    # In this case the code below will have created a default_dir inode for dir1 already.
+                    inode = self.find_inode(path, prefix)
+                except KeyError:
+                    pass
+                else:
+                    self._items[inode] = item
+                    continue
+            segments = prefix + path.split(b'/')
+            parent = 1
+            for segment in segments[:-1]:
+                parent = self._process_inner(segment, parent)
+            self._process_leaf(segments[-1], item, parent, prefix, is_dir, item_inode,
+                               hardlink_masters, strip_components)
+        duration = time.perf_counter() - t0
+        logger.debug('fuse: _process_archive completed in %.1f s for archive %s', duration, archive.name)
+
+    def _process_leaf(self, name, item, parent, prefix, is_dir, item_inode, hardlink_masters, stripped_components):
+        path = item.path
+        del item.path  # save some space
+        hardlink_masters = hardlink_masters or {}
+
+        def file_version(item, path):
+            if 'chunks' in item:
+                file_id = blake2b_128(path)
+                current_version, previous_id = self.versions_index.get(file_id, (0, None))
+
+                contents_id = blake2b_128(b''.join(chunk_id for chunk_id, _, _ in item.chunks))
+
+                if contents_id != previous_id:
+                    current_version += 1
+                    self.versions_index[file_id] = current_version, contents_id
+
+                return current_version
+
+        def make_versioned_name(name, version, add_dir=False):
+            if add_dir:
+                # add intermediate directory with same name as filename
+                path_fname = name.rsplit(b'/', 1)
+                name += b'/' + path_fname[-1]
+            # keep original extension at end to avoid confusing tools
+            name, ext = os.path.splitext(name)
+            version_enc = os.fsencode('.%05d' % version)
+            return name + version_enc + ext
+
+        if 'source' in item and hardlinkable(item.mode):
+            source = os.sep.join(item.source.split(os.sep)[stripped_components:])
+            chunks, link_target = hardlink_masters.get(item.source, (None, source))
+            if link_target:
+                # Hard link was extracted previously, just link
+                link_target = os.fsencode(link_target)
+                if self.versions:
+                    # adjust link target name with version
+                    version = self.file_versions[link_target]
+                    link_target = make_versioned_name(link_target, version, add_dir=True)
+                try:
+                    inode = self.find_inode(link_target, prefix)
+                except KeyError:
+                    logger.warning('Skipping broken hard link: %s -> %s', path, source)
+                    return
+                item = self.get_item(inode)
+                item.nlink = item.get('nlink', 1) + 1
+                self._items[inode] = item
+            elif chunks is not None:
+                # assign chunks to this item, since the item which had the chunks was not extracted
+                item.chunks = chunks
+                inode = item_inode
+                self._items[inode] = item
+                if hardlink_masters:
+                    # Update master entry with extracted item path, so that following hardlinks don't extract twice.
+                    hardlink_masters[item.source] = (None, path)
+        else:
+            inode = item_inode
+
+        if self.versions and not is_dir:
+            parent = self._process_inner(name, parent)
+            enc_path = os.fsencode(path)
+            version = file_version(item, enc_path)
+            if version is not None:
+                # regular file, with contents - maybe a hardlink master
+                name = make_versioned_name(name, version)
+                self.file_versions[enc_path] = version
+
+        self.parent[inode] = parent
+        if name:
+            self.contents[parent][name] = inode
+
+    def _process_inner(self, name, parent_inode):
+        dir = self.contents[parent_inode]
+        if name in dir:
+            inode = dir[name]
+        else:
+            inode = self._create_dir(parent_inode)
+            if name:
+                dir[name] = inode
+        return inode
+
+
+class FuseOperations(llfuse.Operations, FuseBackend):
+    """Export archive as a FUSE filesystem
+    """
+
+    def __init__(self, key, repository, manifest, args, decrypted_repository):
+        llfuse.Operations.__init__(self)
+        FuseBackend.__init__(self, key, manifest, repository, args, decrypted_repository)
+        self.decrypted_repository = decrypted_repository
+        data_cache_capacity = int(os.environ.get('BORG_MOUNT_DATA_CACHE_ENTRIES', os.cpu_count() or 1))
+        logger.debug('mount data cache capacity: %d chunks', data_cache_capacity)
+        self.data_cache = LRUCache(capacity=data_cache_capacity, dispose=lambda _: None)
+        self._last_pos = LRUCache(capacity=FILES, dispose=lambda _: None)
+
     def sig_info_handler(self, sig_no, stack):
         logger.debug('fuse: %d synth inodes, %d edges (%s)',
-                     self._inode_count, len(self.parent),
+                     self.inode_count, len(self.parent),
                      # getsizeof is the size of the dict itself; key and value are two small-ish integers,
                      # which are shared due to code structure (this has been verified).
-                     format_file_size(sys.getsizeof(self.parent) + len(self.parent) * sys.getsizeof(self._inode_count)))
+                     format_file_size(sys.getsizeof(self.parent) + len(self.parent) * sys.getsizeof(self.inode_count)))
         logger.debug('fuse: %d pending archives', len(self.pending_archives))
         logger.debug('fuse: ItemCache %d entries (%d direct, %d indirect), meta-array size %s, direct items size %s',
                      self.cache.direct_items + self.cache.indirect_items, self.cache.direct_items, self.cache.indirect_items,
@@ -340,15 +532,23 @@ class FuseOperations(llfuse.Operations):
         self.umask = pop_option(options, 'umask', 0, 0, int, int_base=8)  # umask is octal, e.g. 222 or 0222
         dir_uid = self.uid_forced if self.uid_forced is not None else self.default_uid
         dir_gid = self.gid_forced if self.gid_forced is not None else self.default_gid
+        dir_user = uid2user(dir_uid)
+        dir_group = gid2group(dir_gid)
+        assert isinstance(dir_user, str)
+        assert isinstance(dir_group, str)
         dir_mode = 0o40755 & ~self.umask
-        self.default_dir = Item(mode=dir_mode, mtime=int(time.time() * 1e9), uid=dir_uid, gid=dir_gid)
+        self.default_dir = Item(mode=dir_mode, mtime=int(time.time() * 1e9),
+                                user=dir_user, group=dir_group, uid=dir_uid, gid=dir_gid)
         self._create_filesystem()
         llfuse.init(self, mountpoint, options)
         if not foreground:
-            old_id, new_id = daemonize()
-            if not isinstance(self.repository_uncached, RemoteRepository):
-                # local repo and the locking process' PID just changed, migrate it:
-                self.repository_uncached.migrate_lock(old_id, new_id)
+            if isinstance(self.repository_uncached, RemoteRepository):
+                daemonize()
+            else:
+                with daemonizing() as (old_id, new_id):
+                    # local repo: the locking process' PID is changing, migrate it:
+                    logger.debug('fuse: mount local repo, going to background: migrating lock.')
+                    self.repository_uncached.migrate_lock(old_id, new_id)
 
         # If the file system crashes, we do not want to umount because in that
         # case the mountpoint suddenly appears to become empty. This can have
@@ -365,147 +565,7 @@ class FuseOperations(llfuse.Operations):
         finally:
             llfuse.close(umount)
 
-    def _create_dir(self, parent, mtime=None):
-        """Create directory
-        """
-        ino = self.allocate_inode()
-        if mtime is not None:
-            self.items[ino] = Item(**self.default_dir.as_dict())
-            self.items[ino].mtime = mtime
-        else:
-            self.items[ino] = self.default_dir
-        self.parent[ino] = parent
-        return ino
-
-    def process_archive(self, archive_name, prefix=[]):
-        """Build FUSE inode hierarchy from archive metadata
-        """
-        self.file_versions = {}  # for versions mode: original path -> version
-        t0 = time.perf_counter()
-        archive = Archive(self.repository_uncached, self.key, self.manifest, archive_name,
-                          consider_part_files=self.args.consider_part_files)
-        strip_components = self.args.strip_components
-        matcher = Archiver.build_matcher(self.args.patterns, self.args.paths)
-        partial_extract = not matcher.empty() or strip_components
-        hardlink_masters = {} if partial_extract else None
-
-        def peek_and_store_hardlink_masters(item, matched):
-            if (partial_extract and not matched and hardlinkable(item.mode) and
-                    item.get('hardlink_master', True) and 'source' not in item):
-                hardlink_masters[item.get('path')] = (item.get('chunks'), None)
-
-        filter = Archiver.build_filter(matcher, peek_and_store_hardlink_masters, strip_components)
-        for item_inode, item in self.cache.iter_archive_items(archive.metadata.items, filter=filter,
-                                                              consider_part_files=self.args.consider_part_files):
-            if strip_components:
-                item.path = os.sep.join(item.path.split(os.sep)[strip_components:])
-            path = os.fsencode(item.path)
-            is_dir = stat.S_ISDIR(item.mode)
-            if is_dir:
-                try:
-                    # This can happen if an archive was created with a command line like
-                    # $ borg create ... dir1/file dir1
-                    # In this case the code below will have created a default_dir inode for dir1 already.
-                    inode = self._find_inode(path, prefix)
-                except KeyError:
-                    pass
-                else:
-                    self.items[inode] = item
-                    continue
-            segments = prefix + path.split(b'/')
-            parent = 1
-            for segment in segments[:-1]:
-                parent = self.process_inner(segment, parent)
-            self.process_leaf(segments[-1], item, parent, prefix, is_dir, item_inode,
-                              hardlink_masters, strip_components)
-        duration = time.perf_counter() - t0
-        logger.debug('fuse: process_archive completed in %.1f s for archive %s', duration, archive.name)
-
-    def process_leaf(self, name, item, parent, prefix, is_dir, item_inode, hardlink_masters, stripped_components):
-        path = item.path
-        del item.path  # save some space
-        hardlink_masters = hardlink_masters or {}
-
-        def file_version(item, path):
-            if 'chunks' in item:
-                file_id = blake2b_128(path)
-                current_version, previous_id = self.versions_index.get(file_id, (0, None))
-
-                chunk_ids = [chunk_id for chunk_id, _, _ in item.chunks]
-                contents_id = blake2b_128(b''.join(chunk_ids))
-
-                if contents_id != previous_id:
-                    current_version += 1
-                    self.versions_index[file_id] = current_version, contents_id
-
-                return current_version
-
-        def make_versioned_name(name, version, add_dir=False):
-            if add_dir:
-                # add intermediate directory with same name as filename
-                path_fname = name.rsplit(b'/', 1)
-                name += b'/' + path_fname[-1]
-            # keep original extension at end to avoid confusing tools
-            name, ext = os.path.splitext(name)
-            version_enc = os.fsencode('.%05d' % version)
-            return name + version_enc + ext
-
-        if 'source' in item and hardlinkable(item.mode):
-            source = os.sep.join(item.source.split(os.sep)[stripped_components:])
-            chunks, link_target = hardlink_masters.get(item.source, (None, source))
-            if link_target:
-                # Hard link was extracted previously, just link
-                link_target = os.fsencode(link_target)
-                if self.versions:
-                    # adjust link target name with version
-                    version = self.file_versions[link_target]
-                    link_target = make_versioned_name(link_target, version, add_dir=True)
-                try:
-                    inode = self._find_inode(link_target, prefix)
-                except KeyError:
-                    logger.warning('Skipping broken hard link: %s -> %s', path, source)
-                    return
-                item = self.get_item(inode)
-                item.nlink = item.get('nlink', 1) + 1
-                self.items[inode] = item
-            elif chunks is not None:
-                # assign chunks to this item, since the item which had the chunks was not extracted
-                item.chunks = chunks
-                inode = item_inode
-                self.items[inode] = item
-                if hardlink_masters:
-                    # Update master entry with extracted item path, so that following hardlinks don't extract twice.
-                    hardlink_masters[item.source] = (None, path)
-        else:
-            inode = item_inode
-
-        if self.versions and not is_dir:
-            parent = self.process_inner(name, parent)
-            enc_path = os.fsencode(path)
-            version = file_version(item, enc_path)
-            if version is not None:
-                # regular file, with contents - maybe a hardlink master
-                name = make_versioned_name(name, version)
-                self.file_versions[enc_path] = version
-
-        self.parent[inode] = parent
-        if name:
-            self.contents[parent][name] = inode
-
-    def process_inner(self, name, parent_inode):
-        dir = self.contents[parent_inode]
-        if name in dir:
-            inode = dir[name]
-        else:
-            inode = self._create_dir(parent_inode)
-            if name:
-                dir[name] = inode
-        return inode
-
-    def allocate_inode(self):
-        self._inode_count += 1
-        return self._inode_count
-
+    @async_wrapper
     def statfs(self, ctx=None):
         stat_ = llfuse.StatvfsData()
         stat_.f_bsize = 512
@@ -516,31 +576,10 @@ class FuseOperations(llfuse.Operations):
         stat_.f_files = 0
         stat_.f_ffree = 0
         stat_.f_favail = 0
-        if hasattr(stat_, 'f_namemax'):  # since llfuse 1.3.0
-            stat_.f_namemax = 255  # == NAME_MAX (depends on archive source OS / FS)
+        stat_.f_namemax = 255  # == NAME_MAX (depends on archive source OS / FS)
         return stat_
 
-    def get_item(self, inode):
-        item = self._inode_cache.get(inode)
-        if item is not None:
-            return item
-        try:
-            # this is a cheap get-from-dictionary operation, no need to cache the result.
-            return self.items[inode]
-        except KeyError:
-            # while self.cache does some internal caching, it has still quite some overhead, so we cache the result.
-            item = self.cache.get(inode)
-            self._inode_cache[inode] = item
-            return item
-
-    def _find_inode(self, path, prefix=[]):
-        segments = prefix + path.split(b'/')
-        inode = 1
-        for segment in segments:
-            inode = self.contents[inode][segment]
-        return inode
-
-    def getattr(self, inode, ctx=None):
+    def _getattr(self, inode, ctx=None):
         item = self.get_item(inode)
         entry = llfuse.EntryAttributes()
         entry.st_ino = inode
@@ -549,32 +588,30 @@ class FuseOperations(llfuse.Operations):
         entry.attr_timeout = 300
         entry.st_mode = item.mode & ~self.umask
         entry.st_nlink = item.get('nlink', 1)
-        entry.st_uid = self.uid_forced if self.uid_forced is not None else item.uid if item.uid >= 0 else self.default_uid
-        entry.st_gid = self.gid_forced if self.gid_forced is not None else item.gid if item.gid >= 0 else self.default_gid
+        entry.st_uid, entry.st_gid = get_item_uid_gid(item, numeric=self.numeric_ids,
+                                                      uid_default=self.default_uid, gid_default=self.default_gid,
+                                                      uid_forced=self.uid_forced, gid_forced=self.gid_forced)
         entry.st_rdev = item.get('rdev', 0)
         entry.st_size = item.get_size()
         entry.st_blksize = 512
         entry.st_blocks = (entry.st_size + entry.st_blksize - 1) // entry.st_blksize
         # note: older archives only have mtime (not atime nor ctime)
-        mtime_ns = item.mtime
-        if have_fuse_xtime_ns:
-            entry.st_mtime_ns = mtime_ns
-            entry.st_atime_ns = item.get('atime', mtime_ns)
-            entry.st_ctime_ns = item.get('ctime', mtime_ns)
-            if have_fuse_birthtime_ns:
-                entry.st_birthtime_ns = item.get('birthtime', mtime_ns)
-        else:
-            entry.st_mtime = mtime_ns / 1e9
-            entry.st_atime = item.get('atime', mtime_ns) / 1e9
-            entry.st_ctime = item.get('ctime', mtime_ns) / 1e9
-            if have_fuse_birthtime:
-                entry.st_birthtime = item.get('birthtime', mtime_ns) / 1e9
+        entry.st_mtime_ns = mtime_ns = item.mtime
+        entry.st_atime_ns = item.get('atime', mtime_ns)
+        entry.st_ctime_ns = item.get('ctime', mtime_ns)
+        entry.st_birthtime_ns = item.get('birthtime', mtime_ns)
         return entry
 
+    @async_wrapper
+    def getattr(self, inode, ctx=None):
+        return self._getattr(inode, ctx=ctx)
+
+    @async_wrapper
     def listxattr(self, inode, ctx=None):
         item = self.get_item(inode)
         return item.get('xattrs', {}).keys()
 
+    @async_wrapper
     def getxattr(self, inode, name, ctx=None):
         item = self.get_item(inode)
         try:
@@ -582,14 +619,9 @@ class FuseOperations(llfuse.Operations):
         except KeyError:
             raise llfuse.FUSEError(llfuse.ENOATTR) from None
 
-    def _load_pending_archive(self, inode):
-        # Check if this is an archive we need to load
-        archive_name = self.pending_archives.pop(inode, None)
-        if archive_name:
-            self.process_archive(archive_name, [os.fsencode(archive_name)])
-
+    @async_wrapper
     def lookup(self, parent_inode, name, ctx=None):
-        self._load_pending_archive(parent_inode)
+        self.check_pending_archive(parent_inode)
         if name == b'.':
             inode = parent_inode
         elif name == b'..':
@@ -598,8 +630,9 @@ class FuseOperations(llfuse.Operations):
             inode = self.contents[parent_inode].get(name)
             if not inode:
                 raise llfuse.FUSEError(errno.ENOENT)
-        return self.getattr(inode)
+        return self._getattr(inode)
 
+    @async_wrapper
     def open(self, inode, flags, ctx=None):
         if not self.allow_damaged_files:
             item = self.get_item(inode)
@@ -610,12 +643,14 @@ class FuseOperations(llfuse.Operations):
                 logger.warning('File has damaged (all-zero) chunks. Try running borg check --repair. '
                                'Mount with allow_damaged_files to read damaged files.')
                 raise llfuse.FUSEError(errno.EIO)
-        return inode
+        return llfuse.FileInfo(fh=inode) if has_pyfuse3 else inode
 
+    @async_wrapper
     def opendir(self, inode, ctx=None):
-        self._load_pending_archive(inode)
+        self.check_pending_archive(inode)
         return inode
 
+    @async_wrapper
     def read(self, fh, offset, size):
         parts = []
         item = self.get_item(fh)
@@ -659,12 +694,25 @@ class FuseOperations(llfuse.Operations):
                 break
         return b''.join(parts)
 
-    def readdir(self, fh, off):
-        entries = [(b'.', fh), (b'..', self.parent[fh])]
-        entries.extend(self.contents[fh].items())
-        for i, (name, inode) in enumerate(entries[off:], off):
-            yield name, self.getattr(inode), i + 1
+    # note: we can't have a generator (with yield) and not a generator (async) in the same method
+    if has_pyfuse3:
+        async def readdir(self, fh, off, token):
+            entries = [(b'.', fh), (b'..', self.parent[fh])]
+            entries.extend(self.contents[fh].items())
+            for i, (name, inode) in enumerate(entries[off:], off):
+                attrs = self._getattr(inode)
+                if not llfuse.readdir_reply(token, name, attrs, i + 1):
+                    break
 
+    else:
+        def readdir(self, fh, off):
+            entries = [(b'.', fh), (b'..', self.parent[fh])]
+            entries.extend(self.contents[fh].items())
+            for i, (name, inode) in enumerate(entries[off:], off):
+                attrs = self._getattr(inode)
+                yield name, attrs, i + 1
+
+    @async_wrapper
     def readlink(self, inode, ctx=None):
         item = self.get_item(inode)
         return os.fsencode(item.source)
